package com.syjy.calculate.service;

import com.googlecode.aviator.AviatorEvaluator;
import com.googlecode.aviator.Expression;
import com.syjy.calculate.entity.CalculateResult;
import com.syjy.calculate.entity.CalculationInfo;
import com.syjy.calculate.entity.CalculateRequest;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;
import java.io.File;
import java.io.IOException;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * 计算准确率、合格率
 *
 * @author zcl
 * @version 1.0
 * @since 2022/8/8 9:30
 */
@Service
@Slf4j
public class AccuracyPassRateCalculateService {

    /**
     * 计算准确率、合格率
     *
     * @param calculateRequest
     * @return
     */
    public CalculateResult calculate(CalculateRequest calculateRequest) {
        // 对细则进行校验
        CalculateResult checkedDataResult = checkDataRules(calculateRequest);
        // 如果细则校验失败，返回细则结果
        if (!String.valueOf(checkedDataResult.get(CalculateResult.CODE_TAG)).equals(CalculateResult.Type.SUCCESS.value())) {
            return checkedDataResult;
        }
        // 获取细则校验过的数据
        List<Map<String, Object>> checkedData = (List<Map<String, Object>>) checkedDataResult.get(CalculateResult.DATA_TAG);
        // 获取jar包所在路径
        String classPath = System.getProperty(CalculateResult.USER_DIR);
        // 获取文件路径
        String fileUrl = classPath + CalculateResult.SLASH + CalculateResult.EXAMPLES + CalculateResult.SLASH + calculateRequest.getProvince() + CalculateResult.UNDERLINE + calculateRequest.getFormulaType() + CalculateResult.UNDERLINE + calculateRequest.getElectricType() + CalculateResult.UNDERLINE + CalculateResult.FORMULA + CalculateResult.AV;
        // 初始化公式并缓存公式

        File file = new File(fileUrl);
        // 判断公式文件是否存在
        if (!file.exists()) {
            // 未匹配到公式文件
            return CalculateResult.error(CalculateResult.MSG_NO_FORMULA + fileUrl);
        }
        Expression exp = null;
        // 初始化公式文件
        try {
            exp = AviatorEvaluator.getInstance().compileScript(fileUrl, true);
        } catch (IOException e) {
            e.printStackTrace();
        }
        // 上下文
        Map<String, Object> env = new HashMap<>();
        // 将需要计算的list数据放入上下文
        env.put(CalculateResult.LIST, checkedData);
        env.put(CalculateResult.RL, calculateRequest.getCap());
        // 执行计算并得出结果
        try {
            // 获取计算结果
            BigDecimal executeResult = new BigDecimal(String.valueOf(exp.execute(env)));
            // 过滤计算结果
            String result = filterResult(executeResult);
            return CalculateResult.success(CalculateResult.MSG_CALCULATE_SUCCESS, result);
        } catch (Exception e) {
            e.printStackTrace();
            return CalculateResult.error(CalculateResult.MSG_CALCULATE_FAIL + e.toString());
        }
    }

    /**
     * 过滤结果 <0=0  >100=100
     *
     * @param result 过滤前的结果
     * @return 过滤后的结果
     */
    private String filterResult(BigDecimal result) {
        //当结果为负数时，说明偏差过大，准确率为0
        if (result.compareTo(BigDecimal.ZERO) == -1 || result.compareTo(BigDecimal.ZERO) == 0) {
            result = BigDecimal.ZERO;
        }
        // 如果结果大于1，则准确率设为100
        if (result.compareTo(BigDecimal.ONE) == 1) {
            result = BigDecimal.ONE;
        }
        // 对数据*100
        result = result.multiply(new BigDecimal(CalculateResult.ONE_HUNDRED));
        // 对数据进行四舍五入
        result = result.setScale(2, BigDecimal.ROUND_HALF_UP);
        // 数据加上百分号
        return result + CalculateResult.PERCENT;
    }

    /**
     * 根据细则过滤数据
     *
     * @param calculateRequest
     * @return
     */
    private CalculateResult checkDataRules(CalculateRequest calculateRequest) {
        // 对入参进行校验
        if (calculateRequest == null) {
            return CalculateResult.error(CalculateResult.MSG_REQUEST_NULL);
        }
        // 公式类型
        String formulaType = calculateRequest.getFormulaType();
        if ("".equals(formulaType) || formulaType == null) {
            return CalculateResult.error(CalculateResult.MSG_FORMULA_TYPE_NULL);
        }
        // 省调
        String province = calculateRequest.getProvince();
        if ("".equals(province) || province == null) {
            return CalculateResult.error(CalculateResult.MSG_PROVINCE_NULL);
        }
        // 场站类型
        String electricType = calculateRequest.getElectricType();
        if ("".equals(electricType) || electricType == null) {
            return CalculateResult.error(CalculateResult.MSG_ELECTRIC_TYPE_NULL);
        }
        // 获取被校验数据
        List<CalculationInfo> calculationInfoList = calculateRequest.getCalculationInfoList();
        // 计算数据为空
        if (calculationInfoList == null || calculationInfoList.size() == 0) {
            return CalculateResult.error(CalculateResult.MSG_CALCULATE_DATA_NULL);
        }
        // 获取jar包所在路径
        String classPath = System.getProperty(CalculateResult.USER_DIR);
        // 获取文件路径
        String fileUrl = classPath + CalculateResult.SLASH + CalculateResult.EXAMPLES + CalculateResult.SLASH + calculateRequest.getProvince() + CalculateResult.UNDERLINE + calculateRequest.getFormulaType() + CalculateResult.UNDERLINE + calculateRequest.getElectricType() + CalculateResult.UNDERLINE + CalculateResult.RULES + CalculateResult.AV;
        // 获取文件
        File file = new File(fileUrl);
        // 将数据转为List Map 格式
        List<Map<String, Object>> calculationInfoListMap = getCalculationInfoList(calculationInfoList);
        // 判断公式文件是否存在
        if (!file.exists()) {
            // 未匹配到细则文件，则不进行细则校验，返回数据
            return CalculateResult.success(calculationInfoListMap);
        }
        List<Map<String, Object>> passList = new ArrayList<>();
        // 初始化细则校验文件
        Expression exp = null;
        try {
            exp = AviatorEvaluator.getInstance().compileScript(fileUrl, true);
        } catch (IOException e) {
            e.printStackTrace();
        }
        // 循环数据，对数据进行校验
        for (Map<String, Object> calculationMap : calculationInfoListMap) {
            // 循环公式，对每条功率信息进行细则处理
            try {
                Map<String,Object> checkResult = (HashMap<String, Object>) exp.execute(calculationMap);
                switch (String.valueOf(checkResult.get(CalculateResult.CODE_TAG))){
                    case CalculateResult.IGNORE :
                        break;
                    case CalculateResult.ERROR :
                        String resultValue = String.valueOf(checkResult.get(CalculateResult.MSG_TAG));
                        return CalculateResult.error(CalculateResult.MSG_DATA_CHECK_ERROR,resultValue);
                default:
                break;
                }
                if (String.valueOf(checkResult.get(CalculateResult.CODE_TAG)).equals(CalculateResult.Type.SUCCESS.value())) {
                    break;
                }
                // 校验成功，将通过的数据放入passList
                passList.add(calculationMap);
            } catch (Exception e) {
                log.error(CalculateResult.MSG_DATA_CHECK_ERROR + e.toString());
                return CalculateResult.error(CalculateResult.MSG_DATA_CHECK_ERROR);
            }
        }
        // 校验成功，返回待计算的数据
        return CalculateResult.success(passList);
    }

    /**
     * 格式化被校验数据
     *
     * @return
     */
    private List<Map<String, Object>> getCalculationInfoList(List<CalculationInfo> calculationInfoList) {
        List<Map<String, Object>> resultList = new ArrayList<>();
        for (CalculationInfo calculationInfo : calculationInfoList) {
            Map<String, Object> map = new HashMap<>();
            map.put(CalculateResult.SJ, calculationInfo.getSj());
            map.put(CalculateResult.YC, calculationInfo.getYc());
            // 如果容量不为空，放入容量字段
            if (calculationInfo.getRl() != null) {
                map.put(CalculateResult.RL, calculationInfo.getRl());
            }
            // 如果时间不为空，放入时间字段
            if (calculationInfo.getTime() != 0) {
                map.put(CalculateResult.TIME, calculationInfo.getTime());
            }
            resultList.add(map);
        }
        return resultList;
    }

}

